from bpx.public import Public

default_public: Public = Public()


def get_volume(filled_orders: list[dict]) -> float | int:
    prices = [float(item['price']) * float(item["quantity"]) for item in filled_orders]
    return sum(prices)


def get_fees(filled_orders: list[dict]) -> float | int:
    fees = [float(item['price']) * float(item["fee"])
            if item["feeSymbol"] != 'USDC' else float(item["fee"]) for item in filled_orders]
    return sum(fees)


def get_approximate_balance_in_usdc(balance: dict,
                                    time_interval: str = "1m",
                                    public: Public = default_public) -> float:
    usdc_symbol = "USDC"
    balance_usdc = 0
    close_price = 0
    for symbol in balance:
        if symbol != usdc_symbol:
            k_lines = public.get_klines(f"{symbol}_{usdc_symbol}", time_interval)
            close_price = float(k_lines[-1]['close'])

        for status in balance[symbol]:
            if symbol == usdc_symbol:
                balance_usdc += float(balance[symbol][status])
            else:
                balance_usdc += float(balance[symbol][status]) * close_price
    return balance_usdc



from enum import Enum


class Blockchain(Enum):
    ETHEREUM = "Ethereum"
    SOLANA = "Solana"
    Polygon = "Polygon"
    BITCOIN = "Bitcoin"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class TimeInterval(Enum):
    ONE_MINUTE = "1m"
    THREE_MINUTES = "3m"
    FIVE_MINUTES = "5m"
    FIFTEEN_MINUTES = "15m"
    THIRTY_MINUTES = "30m"
    ONE_HOUR = "1h"
    TWO_HOURS = "2h"
    FOUR_HOURS = "4h"
    SIX_HOURS = "6h"
    EIGHT_HOURS = "8h"
    TWELVE_HOURS = "12h"
    ONE_DAY = "1d"
    THREE_DAYS = "3d"
    ONE_WEEK = "1w"
    ONE_MONTH = "1month"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class TimeInForce(Enum):
    GTC = "GTC"
    IOC = "IOC"
    FOK = "FOK"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class SelfTradePrevention(Enum):
    REJECT_TAKER = "RejectTaker"
    REJECT_MAKER = "RejectMaker"
    REJECT_BOTH = "RejectBoth"
    ALLOW_ALL = "Allow"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_



from bpx.exceptions import *
from bpx.enums import *


class BasePublic:
    BASE_URL = 'https://api.backpack.exchange/'

    def _endpoint(self, path):
        return f"{self.BASE_URL}{path}"

    def get_assets_url(self):
        return self._endpoint('api/v1/assets')

    def get_markets_url(self):
        return self._endpoint('api/v1/markets')

    def get_ticker_url(self, symbol):
        return self._endpoint(f'api/v1/ticker?symbol={symbol}')

    def get_depth_url(self, symbol):
        return self._endpoint(f'api/v1/depth?symbol={symbol}')

    def get_klines_url(self, symbol, interval, start_time, end_time):
        if start_time < 0 or end_time < 0:
            raise NegativeValueError("start_time and end_time")
        if not TimeInterval.has_value(interval):
            raise InvalidTimeIntervalError(interval)
        url = f'api/v1/klines?symbol={symbol}&interval={interval}'
        if start_time:
            url += f'&startTime={start_time}'
        if end_time:
            url += f'&endTime={end_time}'
        return self._endpoint(url)

    def get_status_url(self):
        return self._endpoint('api/v1/status')

    def get_ping_url(self):
        return self._endpoint('api/v1/ping')

    def get_time_url(self):
        return self._endpoint('api/v1/time')

    def get_recent_trades_url(self, symbol, limit=100):
        if limit < 0 or limit > 1000:
            raise LimitValueError
        return self._endpoint(f'api/v1/trades?symbol={symbol}&limit={limit}')

    def get_history_trades_url(self, symbol, limit=100, offset=0):
        if limit < 0 or limit > 1000:
            raise LimitValueError
        if offset < 0:
            raise NegativeValueError
        return self._endpoint(f'api/v1/trades/history?symbol={symbol}&limit={limit}&offset={offset}')



import aiohttp
from bpx.http_client.http_client import HttpClient
import json
import certifi
import ssl


class AsyncHttpClient(HttpClient):

    def __init__(self, proxy: str = ""):
        self.proxy = proxy

    async def get(self, url, headers=None, params=None):
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        async with aiohttp.ClientSession() as session:
            async with session.get(url, proxy=self.proxy, params=params,
                                   headers=headers, ssl=ssl_context) as response:

                try:
                    return await response.json()
                except json.JSONDecodeError:
                    return await response.text()
                except aiohttp.client_exceptions.ContentTypeError:
                    return await response.text()

    async def post(self, url, headers=None, data=None):
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        async with aiohttp.ClientSession() as session:
            async with session.post(url, proxy=self.proxy, headers=headers,
                                    data=json.dumps(data), ssl=ssl_context) as response:

                try:
                    return await response.json()
                except json.JSONDecodeError:
                    return await response.text()
                except aiohttp.client_exceptions.ContentTypeError:
                    return await response.text()

    async def delete(self, url, headers=None, data=None):
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        async with aiohttp.ClientSession() as session:
            async with session.delete(url, proxy=self.proxy, headers=headers,
                                      data=json.dumps(data), ssl=ssl_context) as response:

                try:
                    return await response.json()
                except json.JSONDecodeError:
                    return await response.text()
                except aiohttp.client_exceptions.ContentTypeError:
                    return await response.text()



from bpx.base.base_public import BasePublic
from bpx.http_client.sync_http_client import SyncHttpClient

default_http_client = SyncHttpClient()


class Public(BasePublic):

    def __init__(self, proxy: dict = None, http_client: SyncHttpClient = default_http_client):
        self.http_client = http_client
        self.http_client.proxies = proxy

    def get_assets(self):
        return self.http_client.get(self.get_assets_url())

    def get_markets(self):
        return self.http_client.get(self.get_markets_url())

    def get_ticker(self, symbol: str):
        return self.http_client.get(self.get_ticker_url(symbol))

    def get_depth(self, symbol: str):
        return self.http_client.get(self.get_depth_url(symbol))

    def get_klines(self, symbol: str, interval: str, start_time=0, end_time=0):
        return self.http_client.get(self.get_klines_url(symbol, interval, start_time, end_time))

    def get_status(self):
        return self.http_client.get(self.get_status_url())

    def get_ping(self):
        return self.http_client.get(self.get_ping_url())

    def get_time(self):
        return self.http_client.get(self.get_time_url())

    def get_recent_trades(self, symbol: str, limit=100):
        return self.http_client.get(self.get_recent_trades_url(symbol, limit))

    def get_history_trades(self, symbol: str, limit=100, offset=0):
        return self.http_client.get(self.get_history_trades_url(symbol, limit, offset))



from cryptography.hazmat.primitives.asymmetric import ed25519
import base64
from time import time
from bpx.exceptions import *
from bpx.enums import *


class BaseAccount:
    BPX_API_URL = 'https://api.backpack.exchange/'

    def __init__(self,
                 public_key: str,
                 secret_key: str,
                 window: int,
                 debug: bool):

        self.private_key = ed25519.Ed25519PrivateKey.from_private_bytes(
            base64.b64decode(secret_key)
        )
        self.public_key = public_key
        self.window = window
        self.debug = debug

    def get_balances(self, window: int):
        headers = self._headers({}, 'balanceQuery', window=window)
        url = self.BPX_API_URL + 'api/v1/capital'
        return url, headers

    def deposits(self, limit: int, offset: int, window: int):
        if limit > 1000 or limit < 0:
            raise LimitValueError
        if offset < 0:
            raise NegativeValueError(offset)
        params = {
            'limit': limit,
            'offset': offset,
        }
        headers = self._headers(params, 'depositQueryAll', window=window)
        url = self.BPX_API_URL + 'wapi/v1/capital/deposits'
        return url, headers, params

    def get_deposit_address(self, blockchain: str, window: int):
        if not Blockchain.has_value(blockchain):
            raise InvalidBlockchainValue(blockchain)
        params = {'blockchain': blockchain}
        headers = self._headers(params, 'depositAddressQuery', window=window)
        url = self.BPX_API_URL + 'wapi/v1/capital/deposit/address'
        return url, headers, params

    def get_withdrawals(self, limit: int, offset: int, window: int):
        if limit > 1000 or limit < 0:
            raise LimitValueError
        if offset < 0:
            raise NegativeValueError(offset)
        params = {'limit': limit, 'offset': offset}
        headers = self._headers(params, 'withdrawalQueryAll', window=window)
        url = self.BPX_API_URL + 'wapi/v1/capital/withdrawals'
        return url, headers, params

    def withdrawal(self, address: str,
                   symbol: str,
                   blockchain: str,
                   quantity: str,
                   window: int):
        if not Blockchain.has_value(blockchain):
            raise InvalidBlockchainValue(blockchain)
        params = {
            'address': address,
            'blockchain': blockchain,
            'quantity': quantity,
            'symbol': symbol,
        }
        headers = self._headers(params, 'withdraw', window=window)
        url = self.BPX_API_URL + 'wapi/v1/capital/withdrawals'
        return url, headers, params

    def get_order_history_query(self, symbol: str, limit: int, offset: int, window: int):
        if limit > 1000 or limit < 0:
            raise LimitValueError
        if offset < 0:
            raise NegativeValueError(offset)
        params = {'symbol': symbol,
                  'limit': limit,
                  'offset': offset}
        headers = self._headers(params, 'orderHistoryQueryAll', window=window)
        url = self.BPX_API_URL + 'wapi/v1/history/orders'
        return url, headers, params

    def get_fill_history_query(self, symbol: str,
                               limit: int,
                               offset: int,
                               __from: int,
                               to: int,
                               window: int):
        if limit > 1000 or limit < 0:
            raise LimitValueError
        if offset < 0:
            raise NegativeValueError(offset)
        params = {
            'symbol': symbol,
            'limit': limit,
            'offset': offset,
        }
        if __from:
            if __from < 0:
                raise NegativeValueError(__from)
            else:
                params['from'] = __from
        if to:
            if to < 0:
                raise NegativeValueError(to)
            else:
                params['to'] = to
        headers = self._headers(params, 'fillHistoryQueryAll', window=window)
        url = self.BPX_API_URL + 'wapi/v1/history/fills'
        return url, headers, params

    def get_open_order(self, symbol: str,
                       order_id: str,
                       client_id: int,
                       window: int):

        params = {'symbol': symbol}
        if order_id:
            params['orderId'] = order_id
        if client_id:
            params['clientId'] = client_id
        headers = self._headers(params, 'orderQuery', window=window)
        url = self.BPX_API_URL + 'api/v1/order'
        return url, headers, params

    def execute_order(self, symbol: str,
                      side: str,
                      order_type: str,
                      time_in_force: str,
                      quantity: float,
                      price: float,
                      trigger_price: float,
                      self_trade_prevention: str,
                      quote_quantity: float,
                      client_id: int,
                      post_only: bool,
                      window: int):
        if not SelfTradePrevention.has_value(self_trade_prevention):
            raise InvalidSelfTradePreventionError(self_trade_prevention)
        params = {
            'symbol': symbol,
            'side': side,
            'orderType': order_type,
            'quantity': quantity,
            'selfTradePrevention': self_trade_prevention,
        }
        if order_type != "Market":
            if price:
                params['price'] = price
        if trigger_price:
            params['triggerPrice'] = trigger_price
        if quote_quantity:
            params['quoteQuantity'] = quote_quantity
        if post_only:
            params['postOnly'] = True
        else:
            if TimeInForce.has_value(time_in_force):
                params['timeInForce'] = time_in_force
            else:
                raise InvalidTimeInForceValue(time_in_force)
        if client_id:
            params['clientId'] = client_id
        headers = self._headers(params, 'orderExecute', window=window)
        url = self.BPX_API_URL + 'api/v1/order'
        return url, headers, params

    def cancel_order(self, symbol: str,
                     order_id: str,
                     client_id: int,
                     window: int):
        params = {'symbol': symbol}
        if order_id:
            params['orderId'] = order_id
        if client_id:
            params['clientId'] = client_id
        headers = self._headers(params, 'orderCancel', window=window)
        url = self.BPX_API_URL + 'api/v1/order'
        return url, headers, params

    def get_open_orders(self, symbol: str = None, window: int = None):
        params = {'symbol': symbol}
        headers = self._headers(params, 'orderQueryAll', window=window)
        url = self.BPX_API_URL + 'api/v1/orders'
        return url, headers, params

    def cancel_all_orders(self, symbol: str, window: int = None):
        params = {'symbol': symbol}
        headers = self._headers(params, 'orderCancelAll', window=window)
        url = self.BPX_API_URL + 'api/v1/orders'
        return url, headers, params

    def _headers(self,
                 params: dict,
                 instruction: str,
                 window: int) -> dict:
        window = self.window if window is None else window
        timestamp = int(time() * 1e3)
        encoded_signature = self._sign(params, instruction, timestamp, window)
        headers = {
            "X-API-Key": self.public_key,
            "X-Signature": encoded_signature,
            "X-Timestamp": str(timestamp),
            "X-Window": str(window),
            "Content-Type": "application/json; charset=utf-8",
        }
        if self.debug:
            print(headers)
        return headers

    def _sign(self, params: dict, instruction: str, timestamp: int, window: int):
        sign_str = f"instruction={instruction}"
        sorted_params = "&".join(
            f"{key}={value}" for key, value in sorted(params.items())
        )
        if sorted_params:
            sign_str += "&" + sorted_params
        sign_str += f"&timestamp={timestamp}&window={window}"
        if self.debug:
            print(sign_str)
        signature_bytes = self.private_key.sign(sign_str.encode())
        encoded_signature = base64.b64encode(signature_bytes).decode()
        return encoded_signature



from bpx.base.base_account import BaseAccount
from bpx.http_client.async_http_client import AsyncHttpClient

default_http_client = AsyncHttpClient()


class Account(BaseAccount):
    def __init__(self,
                 public_key: str,
                 secret_key: str,
                 window: int = 5000,
                 proxy: str = None,
                 debug: bool = False,
                 http_client: AsyncHttpClient = default_http_client
                 ):

        super().__init__(public_key,
                         secret_key,
                         window,
                         debug)
        self.http_client = http_client
        self.http_client.proxy = proxy

    async def get_balances(self, window: int = None):
        url, headers = super().get_balances(window)
        return await self.http_client.get(url, headers=headers)

    async def deposits(self, limit: int = 100, offset: int = 0, window: int = None):
        url, headers, params = super().deposits(limit, offset, window)
        return await self.http_client.get(url, headers=headers, params=params)

    async def get_deposit_address(self, blockchain: str, window: int = None):
        url, headers, params = super().get_deposit_address(blockchain, window)
        return await self.http_client.get(url, headers=headers, params=params)

    async def get_withdrawals(self, limit: int = 100, offset: int = 0, window: int = None):
        url, headers, params = super().get_withdrawals(limit, offset, window)
        return await self.http_client.get(url, headers=headers, params=params)

    async def withdrawal(self, address: str,
                   symbol: str,
                   blockchain: str,
                   quantity: str,
                   window: int = None):
        url, headers, params = super().withdrawal(address, symbol, blockchain, quantity, window)
        return await self.http_client.post(url, headers=headers, data=params)

    async def get_order_history_query(self, symbol: str, limit: int = 100, offset: int = 0, window: int = None):
        url, headers, params = super().get_order_history_query(symbol, limit, offset, window)
        return await self.http_client.get(url, headers=headers, params=params)

    async def get_fill_history_query(self, symbol: str,
                               limit: int = 100,
                               offset: int = 0,
                               __from: int = None,
                               to: int = None,
                               window: int = None):
        url, headers, params = super().get_fill_history_query(symbol,
                                                         limit,
                                                         offset,
                                                         __from,
                                                         to,
                                                         window)
        return await self.http_client.get(url, headers=headers, params=params)

    async def get_open_order(self, symbol: str,
                       order_id: str = None,
                       client_id: int = None,
                       window: int = None):
        url, headers, params = super().get_open_order(symbol, order_id, client_id, window)
        return await self.http_client.get(url, headers=headers, params=params)

    async def execute_order(self, symbol: str,
                      side: str,
                      order_type: str,
                      quantity: float,
                      time_in_force: str = None,
                      price: float = 0,
                      trigger_price: float = 0,
                      self_trade_prevention: str = "RejectBoth",
                      quote_quantity: float = None,
                      client_id: int = None,
                      post_only: bool = None,
                      window: int = None):

        url, headers, params = super().execute_order(symbol, side, order_type,
                                                time_in_force, quantity,
                                                price, trigger_price,
                                                self_trade_prevention, quote_quantity,
                                                client_id, post_only, window)
        return await self.http_client.post(url, headers=headers, data=params)

    async def cancel_order(self, symbol: str,
                     order_id: str = None,
                     client_id: int = None,
                     window: int = None):
        url, headers, params = super().cancel_order(symbol, order_id, client_id, window)
        return await self.http_client.delete(url, headers=headers, data=params)

    async def get_open_orders(self, symbol: str = None, window: int = None):
        url, headers, params = super().get_open_orders(symbol, window)
        return await self.http_client.get(url, headers=headers, params=params)

    async def cancel_all_orders(self, symbol: str, window: int = None):
        url, headers, params = super().cancel_all_orders(symbol, window)
        return await self.http_client.delete(url, headers=headers, data=params)





class LimitValueError(Exception):
    """Exception when value is out of maximal allowed"""
    MIN = 0
    MAX = 1000

    def __init__(self):
        super().__init__(f"Limit value is out of allowed minimal {self.MIN} and maximal {self.MAX}.")


class NegativeValueError(Exception):
    """Exception when value is negative, when it's not expected to be"""
    def __init__(self, value):
        super().__init__(f"Value {value} can't be negative")


class InvalidBlockchainValue(Exception):
    """Exception when blockchain value is invalid"""
    def __init__(self, value):
        super().__init__(f"Blockchain value {value} is not supported\n"
                         f"https://docs.backpack.exchange/#tag/Capital/operation/get_deposits")


class InvalidTimeIntervalError(Exception):
    def __init__(self, invalid_value):
        self.invalid_value = invalid_value
        documentation_url = "https://docs.backpack.exchange/#tag/Markets/operation/get_klines"
        message = (f"{invalid_value} is not a valid time interval.\n"
                   f"See the documentation for more details: {documentation_url}")
        super().__init__(message)


class InvalidTimeInForceValue(Exception):
    def __init__(self, value):
        documentation_url = "https://support.backpack.exchange/en/articles/543681"
        super().__init__(f"No such timeInForce value {value}\n"
                         f"See the documentation for more details: {documentation_url}")


class InvalidSelfTradePreventionError(Exception):
    def __init__(self, value):
        documentation_url = "https://support.backpack.exchange/en/articles/543681"
        super().__init__(f"No such selfTradePrevention value {value}\n"
                         f"See the documentation for more details: {documentation_url}")



import abc


class HttpClient(abc.ABC):
    @abc.abstractmethod
    def get(self, url, headers=None, params=None):
        """Perform a GET request."""
        pass

    @abc.abstractmethod
    def post(self, url, headers=None, data=None):
        """Perform a POST request."""
        pass

    @abc.abstractmethod
    def delete(self, url, headers=None):
        """Perform a DELETE request"""
        pass



import requests
from bpx.http_client.http_client import HttpClient
import json


class SyncHttpClient(HttpClient):
    def __init__(self, proxies: dict = None):
        self.proxies = proxies

    def get(self, url, headers=None, params=None):
        response = requests.get(url=url, proxies=self.proxies, headers=headers, params=params)
        try:
            return response.json()
        except json.JSONDecodeError:
            return response.text

    def post(self, url, headers=None, data=None):
        response = requests.post(url=url, proxies=self.proxies, headers=headers, data=json.dumps(data))
        try:
            return response.json()
        except json.JSONDecodeError:
            return response.text

    def delete(self, url, headers=None, data=None):
        response = requests.delete(url, proxies=self.proxies, headers=headers, data=json.dumps(data))
        try:
            return response.json()
        except json.JSONDecodeError:
            return response.text
